import os

import httpx

from exchange import Message, Tool
from exchange.content import Text, ToolResult, ToolUse
from exchange.providers.base import Provider, Usage
from tenacity import retry, wait_fixed, stop_after_attempt
from exchange.providers.utils import retry_if_status, raise_for_status
from exchange.langfuse_wrapper import observe_wrapper

ANTHROPIC_HOST = "https://api.anthropic.com/v1/messages"

retry_procedure = retry(
    wait=wait_fixed(2),
    stop=stop_after_attempt(2),
    retry=retry_if_status(codes=[429], above=500),
    reraise=True,
)


class AnthropicProvider(Provider):
    """Provides chat completions for models hosted directly by Anthropic."""

    PROVIDER_NAME = "anthropic"
    REQUIRED_ENV_VARS = ["ANTHROPIC_API_KEY"]

    def __init__(self, client: httpx.Client) -> None:
        self.client = client

    @classmethod
    def from_env(cls: type["AnthropicProvider"]) -> "AnthropicProvider":
        cls.check_env_vars()
        url = os.environ.get("ANTHROPIC_HOST", ANTHROPIC_HOST)
        key = os.environ.get("ANTHROPIC_API_KEY")
        client = httpx.Client(
            base_url=url,
            headers={
                "x-api-key": key,
                "content-type": "application/json",
                "anthropic-version": "2023-06-01",
            },
            timeout=httpx.Timeout(60 * 10),
        )
        return cls(client)

    @staticmethod
    def get_usage(data: dict) -> Usage:  # noqa: ANN401
        usage = data.get("usage")
        input_tokens = usage.get("input_tokens")
        output_tokens = usage.get("output_tokens")
        total_tokens = usage.get("total_tokens")

        if total_tokens is None and input_tokens is not None and output_tokens is not None:
            total_tokens = input_tokens + output_tokens

        return Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )

    @staticmethod
    def anthropic_response_to_message(response: dict) -> Message:
        content_blocks = response.get("content", [])
        content = []
        for block in content_blocks:
            if block["type"] == "text":
                content.append(Text(text=block["text"]))
            elif block["type"] == "tool_use":
                content.append(
                    ToolUse(
                        id=block["id"],
                        name=block["name"],
                        parameters=block["input"],
                    )
                )
        return Message(role="assistant", content=content)

    @staticmethod
    def tools_to_anthropic_spec(tools: tuple[Tool, ...]) -> list[dict[str, any]]:
        return [
            {
                "name": tool.name,
                "description": tool.description or "",
                "input_schema": tool.parameters,
            }
            for tool in tools
        ]

    @staticmethod
    def messages_to_anthropic_spec(messages: list[Message]) -> list[dict[str, any]]:
        messages_spec = []
        # if messages is empty - just make a default
        for message in messages:
            converted = {"role": message.role}
            for content in message.content:
                if isinstance(content, Text):
                    converted["content"] = [{"type": "text", "text": content.text}]
                elif isinstance(content, ToolUse):
                    converted.setdefault("content", []).append(
                        {
                            "type": "tool_use",
                            "id": content.id,
                            "name": content.name,
                            "input": content.parameters,
                        }
                    )
                elif isinstance(content, ToolResult):
                    converted.setdefault("content", []).append(
                        {
                            "type": "tool_result",
                            "tool_use_id": content.tool_use_id,
                            "content": content.output,
                        }
                    )
            messages_spec.append(converted)
        if len(messages_spec) == 0:
            converted = {
                "role": "user",
                "content": [{"type": "text", "text": "Ignore"}],
            }
            messages_spec.append(converted)
        return messages_spec

    @observe_wrapper(as_type="generation")
    def complete(
        self,
        model: str,
        system: str,
        messages: list[Message],
        tools: list[Tool] = None,
        **kwargs: dict[str, any],
    ) -> tuple[Message, Usage]:
        if tools is None:
            tools = []
        tools_set = set()
        unique_tools = []
        for tool in tools:
            if tool.name not in tools_set:
                unique_tools.append(tool)
                tools_set.add(tool.name)

        payload = dict(
            system=system,
            model=model,
            max_tokens=4096,
            messages=self.messages_to_anthropic_spec(messages),
            tools=self.tools_to_anthropic_spec(tuple(unique_tools)),
            **kwargs,
        )
        payload = {k: v for k, v in payload.items() if v}

        response = self._post(payload)
        message = self.anthropic_response_to_message(response)
        usage = self.get_usage(response)

        return message, usage

    @retry_procedure
    def _post(self, payload: dict) -> httpx.Response:
        response = self.client.post(ANTHROPIC_HOST, json=payload)
        return raise_for_status(response).json()



import pytest
from exchange.invalid_choice_error import InvalidChoiceError
from exchange.providers import get_provider


def test_get_provider_valid():
    provider_name = "openai"
    provider = get_provider(provider_name)
    assert provider.__name__ == "OpenAiProvider"


def test_get_provider_throw_error_for_unknown_provider():
    with pytest.raises(InvalidChoiceError) as error:
        get_provider("nonexistent")
    assert error.value.attribute_name == "provider"
    assert error.value.attribute_value == "nonexistent"
    assert "openai" in error.value.available_values
    assert "openai" in error.value.message



from functools import cache

from exchange.invalid_choice_error import InvalidChoiceError
from exchange.providers.anthropic import AnthropicProvider  # noqa
from exchange.providers.base import Provider, Usage  # noqa
from exchange.providers.databricks import DatabricksProvider  # noqa
from exchange.providers.openai import OpenAiProvider  # noqa
from exchange.providers.ollama import OllamaProvider  # noqa
from exchange.providers.groq import GroqProvider  # noqa
from exchange.providers.azure import AzureProvider  # noqa
from exchange.providers.google import GoogleProvider  # noqa

from exchange.utils import load_plugins


@cache
def get_provider(name: str) -> type[Provider]:
    providers = load_plugins(group="exchange.provider")
    if name not in providers:
        raise InvalidChoiceError("provider", name, providers.keys())
    return providers[name]



from exchange.invalid_choice_error import InvalidChoiceError


def test_load_invalid_choice_error():
    attribute_name = "moderator"
    attribute_value = "not_exist"
    available_values = ["truncate", "summarizer"]
    error = InvalidChoiceError(attribute_name, attribute_value, available_values)

    assert error.attribute_name == attribute_name
    assert error.attribute_value == attribute_value
    assert error.attribute_value == attribute_value
    assert error.message == "Unknown moderator: not_exist. Available moderators: truncate, summarizer"



import os

import httpx

from exchange.providers.openai import OpenAiProvider

OLLAMA_HOST = "http://localhost:11434/"
OLLAMA_MODEL = "qwen2.5"


class OllamaProvider(OpenAiProvider):
    """Provides chat completions for models hosted by Ollama."""

    __doc__ += f"""Here's an example profile configuration to try:

First run: ollama pull qwen2.5, then use this profile:

ollama:
  provider: ollama
  processor: {OLLAMA_MODEL}
  accelerator: {OLLAMA_MODEL}
  moderator: truncate
  toolkits:
  - name: developer
    requires: {{}}
"""

    def __init__(self, client: httpx.Client) -> None:
        print("PLEASE NOTE: the ollama provider is experimental, use with care")
        super().__init__(client)

    @classmethod
    def from_env(cls: type["OllamaProvider"]) -> "OllamaProvider":
        ollama_url = os.environ.get("OLLAMA_HOST", OLLAMA_HOST)
        timeout = httpx.Timeout(60 * 10)

        # from_env is expected to fail if required ENV variables are not
        # available. Since this provider can run with defaults, we substitute
        # an Ollama health check (GET /) to determine if the service is ok.
        httpx.get(ollama_url, timeout=timeout)

        # When served by Ollama, the OpenAI API is available at the path "v1/".
        client = httpx.Client(base_url=ollama_url + "v1/", timeout=timeout)
        return cls(client)



import os

import httpx

from exchange import Message, Tool
from exchange.content import Text, ToolResult, ToolUse
from exchange.providers.base import Provider, Usage
from tenacity import retry, wait_fixed, stop_after_attempt
from exchange.providers.utils import raise_for_status, retry_if_status, encode_image
from exchange.langfuse_wrapper import observe_wrapper


GOOGLE_HOST = "https://generativelanguage.googleapis.com/v1beta"

retry_procedure = retry(
    wait=wait_fixed(2),
    stop=stop_after_attempt(2),
    retry=retry_if_status(codes=[429], above=500),
    reraise=True,
)


class GoogleProvider(Provider):
    """Provides chat completions for models hosted by Google, including Gemini and other experimental models."""

    PROVIDER_NAME = "google"
    REQUIRED_ENV_VARS = ["GOOGLE_API_KEY"]
    instructions_url = "https://ai.google.dev/gemini-api/docs/api-key"

    def __init__(self, client: httpx.Client) -> None:
        self.client = client

    @classmethod
    def from_env(cls: type["GoogleProvider"]) -> "GoogleProvider":
        cls.check_env_vars(cls.instructions_url)
        url = os.environ.get("GOOGLE_HOST", GOOGLE_HOST)
        key = os.environ.get("GOOGLE_API_KEY")
        client = httpx.Client(
            base_url=url,
            headers={
                "Content-Type": "application/json",
            },
            params={"key": key},
            timeout=httpx.Timeout(60 * 10),
        )
        return cls(client)

    @staticmethod
    def get_usage(data: dict) -> Usage:  # noqa: ANN401
        usage = data.get("usageMetadata")
        input_tokens = usage.get("promptTokenCount")
        output_tokens = usage.get("candidatesTokenCount")
        total_tokens = usage.get("totalTokenCount")

        if total_tokens is None and input_tokens is not None and output_tokens is not None:
            total_tokens = input_tokens + output_tokens

        return Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )

    @staticmethod
    def google_response_to_message(response: dict) -> Message:
        candidates = response.get("candidates", [])
        if candidates:
            # Only use first candidate for now
            candidate = candidates[0]
            content_parts = candidate.get("content", {}).get("parts", [])
            content = []
            for part in content_parts:
                if "text" in part:
                    content.append(Text(text=part["text"]))
                elif "functionCall" in part:
                    content.append(
                        ToolUse(
                            id=part["functionCall"].get("name", ""),
                            name=part["functionCall"].get("name", ""),
                            parameters=part["functionCall"].get("args", {}),
                        )
                    )
            return Message(role="assistant", content=content)

        # If no valid candidates were found, return an empty message
        return Message(role="assistant", content=[])

    @staticmethod
    def tools_to_google_spec(tools: tuple[Tool, ...]) -> dict[str, list[dict[str, any]]]:
        if not tools:
            return {}
        converted_tools = []
        for tool in tools:
            converted_tool: dict[str, any] = {
                "name": tool.name,
                "description": tool.description or "",
            }
            if tool.parameters["properties"]:
                converted_tool["parameters"] = tool.parameters
            converted_tools.append(converted_tool)
        return {"functionDeclarations": converted_tools}

    @staticmethod
    def messages_to_google_spec(messages: list[Message]) -> list[dict[str, any]]:
        messages_spec = []
        for message in messages:
            role = "user" if message.role == "user" else "model"
            converted = {"role": role, "parts": []}
            for content in message.content:
                if isinstance(content, Text):
                    converted["parts"].append({"text": content.text})
                elif isinstance(content, ToolUse):
                    converted["parts"].append({"functionCall": {"name": content.name, "args": content.parameters}})
                elif isinstance(content, ToolResult):
                    if content.output.startswith('"image:'):
                        image_path = content.output.replace('"image:', "").replace('"', "")
                        converted["parts"].append(
                            {
                                "inline_data": {
                                    "mime_type": "image/png",
                                    "data": f"{encode_image(image_path)}",
                                }
                            }
                        )
                    else:
                        converted["parts"].append(
                            {"functionResponse": {"name": content.tool_use_id, "response": {"content": content.output}}}
                        )
            messages_spec.append(converted)

        if not messages_spec:
            messages_spec.append({"role": "user", "parts": [{"text": "Ignore"}]})

        return messages_spec

    @observe_wrapper(as_type="generation")
    def complete(
        self,
        model: str,
        system: str,
        messages: list[Message],
        tools: list[Tool] = None,
        **kwargs: dict[str, any],
    ) -> tuple[Message, Usage]:
        tools_set = set()
        unique_tools = []
        for tool in tools:
            if tool.name not in tools_set:
                unique_tools.append(tool)
                tools_set.add(tool.name)

        payload = dict(
            system_instruction={"parts": [{"text": system}]},
            contents=self.messages_to_google_spec(messages),
            tools=self.tools_to_google_spec(tuple(unique_tools)),
            **kwargs,
        )
        payload = {k: v for k, v in payload.items() if v}
        response = self._post(payload, model)
        message = self.google_response_to_message(response)
        usage = self.get_usage(response)
        return message, usage

    @retry_procedure
    def _post(self, payload: dict, model: str) -> httpx.Response:
        response = self.client.post("models/" + model + ":generateContent", json=payload)
        return raise_for_status(response).json()



import logging
import os
from unittest.mock import patch

import pytest
from exchange.content import Text, ToolResult, ToolUse
from exchange.message import Message
from exchange.providers.base import MissingProviderEnvVariableError
from exchange.providers.bedrock import BedrockProvider
from exchange.tool import Tool

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "env_var_name",
    [
        "AWS_ACCESS_KEY_ID",
        "AWS_SECRET_ACCESS_KEY",
        "AWS_SESSION_TOKEN",
    ],
)
def test_from_env_throw_error_when_missing_env_var(env_var_name):
    with patch.dict(
        os.environ,
        {
            "AWS_ACCESS_KEY_ID": "test_access_key_id",
            "AWS_SECRET_ACCESS_KEY": "test_secret_access_key",
            "AWS_SESSION_TOKEN": "test_session_token",
        },
        clear=True,
    ):
        os.environ.pop(env_var_name)
        with pytest.raises(MissingProviderEnvVariableError) as context:
            BedrockProvider.from_env()
        assert context.value.provider == "bedrock"
        assert context.value.env_variable == env_var_name
        assert context.value.message == f"Missing environment variable: {env_var_name} for provider bedrock."


@pytest.fixture
@patch.dict(
    os.environ,
    {
        "AWS_REGION": "us-east-1",
        "AWS_ACCESS_KEY_ID": "fake-access-key",
        "AWS_SECRET_ACCESS_KEY": "fake-secret-key",
        "AWS_SESSION_TOKEN": "fake-session-token",
    },
)
def bedrock_provider():
    return BedrockProvider.from_env()


@patch("time.time", return_value=1624250000)
def test_sign_and_get_headers(mock_time, bedrock_provider):
    # Create sample values
    method = "POST"
    url = "https://bedrock-runtime.us-east-1.amazonaws.com/some/path"
    payload = {"key": "value"}
    service = "bedrock"
    # Generate headers
    headers = bedrock_provider.client.sign_and_get_headers(
        method,
        url,
        payload,
        service,
    )
    # Assert that headers contain expected keys
    assert "Authorization" in headers
    assert "Content-Type" in headers
    assert "X-Amz-date" in headers
    assert "x-amz-content-sha256" in headers
    assert "X-Amz-Security-Token" in headers


@patch("httpx.Client.post")
def test_complete(mock_post, bedrock_provider):
    # Mocked response from the server
    mock_response = {
        "output": {"message": {"role": "assistant", "content": [{"text": "Hello, world!"}]}},
        "usage": {"inputTokens": 10, "outputTokens": 15, "totalTokens": 25},
    }
    mock_post.return_value.json.return_value = mock_response

    model = "test-model"
    system = "You are a helpful assistant."
    messages = [Message.user("Hello")]
    tools = ()

    reply_message, reply_usage = bedrock_provider.complete(model=model, system=system, messages=messages, tools=tools)

    # Assertions for reply message
    assert reply_message.content[0].text == "Hello, world!"
    assert reply_usage.total_tokens == 25


def test_message_to_bedrock_spec_text(bedrock_provider):
    message = Message(role="user", content=[Text("Hello, world!")])
    expected = {"role": "user", "content": [{"text": "Hello, world!"}]}
    assert bedrock_provider.message_to_bedrock_spec(message) == expected


def test_message_to_bedrock_spec_tool_use(bedrock_provider):
    tool_use = ToolUse(id="tool-1", name="WordCount", parameters={"text": "Hello, world!"})
    message = Message(role="assistant", content=[tool_use])
    expected = {
        "role": "assistant",
        "content": [
            {
                "toolUse": {
                    "toolUseId": "tool-1",
                    "name": "WordCount",
                    "input": {"text": "Hello, world!"},
                }
            }
        ],
    }
    assert bedrock_provider.message_to_bedrock_spec(message) == expected


def test_message_to_bedrock_spec_tool_result(bedrock_provider):
    message = Message(
        role="assistant",
        content=[ToolUse(id="tool-1", name="WordCount", parameters={"text": "Hello, world!"})],
    )
    expected = {
        "role": "assistant",
        "content": [
            {
                "toolUse": {
                    "toolUseId": "tool-1",
                    "name": "WordCount",
                    "input": {"text": "Hello, world!"},
                }
            }
        ],
    }
    assert bedrock_provider.message_to_bedrock_spec(message) == expected


def test_message_to_bedrock_spec_tool_result_text(bedrock_provider):
    tool_result = ToolResult(tool_use_id="tool-1", output="Error occurred", is_error=True)
    message = Message(role="user", content=[tool_result])
    expected = {
        "role": "user",
        "content": [
            {
                "toolResult": {
                    "toolUseId": "tool-1",
                    "content": [{"text": "Error occurred"}],
                    "status": "error",
                }
            }
        ],
    }
    assert bedrock_provider.message_to_bedrock_spec(message) == expected


def test_message_to_bedrock_spec_invalid(bedrock_provider):
    with pytest.raises(Exception):
        bedrock_provider.message_to_bedrock_spec(Message(role="user", content=[]))


def test_response_to_message_text(bedrock_provider):
    response = {"role": "user", "content": [{"text": "Hello, world!"}]}
    message = bedrock_provider.response_to_message(response)
    assert message.role == "user"
    assert message.content[0].text == "Hello, world!"


def test_response_to_message_tool_use(bedrock_provider):
    response = {
        "role": "assistant",
        "content": [
            {
                "toolUse": {
                    "toolUseId": "tool-1",
                    "name": "WordCount",
                    "input": {"text": "Hello, world!"},
                }
            }
        ],
    }
    message = bedrock_provider.response_to_message(response)
    assert message.role == "assistant"
    assert message.content[0].name == "WordCount"
    assert message.content[0].parameters == {"text": "Hello, world!"}


def test_response_to_message_tool_result(bedrock_provider):
    response = {
        "role": "user",
        "content": [
            {
                "toolResult": {
                    "toolResultId": "tool-1",
                    "content": [{"json": {"result": 2}}],
                }
            }
        ],
    }
    message = bedrock_provider.response_to_message(response)
    assert message.role == "user"
    assert message.content[0].tool_use_id == "tool-1"
    assert message.content[0].output == {"result": 2}


def test_response_to_message_invalid(bedrock_provider):
    with pytest.raises(Exception):
        bedrock_provider.response_to_message({})


def test_tools_to_bedrock_spec(bedrock_provider):
    def word_count(text: str):
        return len(text.split())

    tool = Tool(
        name="WordCount",
        description="Counts words.",
        parameters={"text": "string"},
        function=word_count,
    )
    expected = {
        "tools": [
            {
                "toolSpec": {
                    "name": "WordCount",
                    "description": "Counts words.",
                    "inputSchema": {"json": {"text": "string"}},
                }
            }
        ]
    }
    assert bedrock_provider.tools_to_bedrock_spec((tool,)) == expected


def test_tools_to_bedrock_spec_duplicate(bedrock_provider):
    def word_count(text: str):
        return len(text.split())

    tool = Tool(
        name="WordCount",
        description="Counts words.",
        parameters={"text": "string"},
        function=word_count,
    )
    tool_duplicate = Tool(
        name="WordCount",
        description="Counts words.",
        parameters={"text": "string"},
        function=word_count,
    )
    tools = bedrock_provider.tools_to_bedrock_spec((tool, tool_duplicate))
    assert set(tool["toolSpec"]["name"] for tool in tools["tools"]) == {"WordCount"}



import base64
import json
import re
from typing import Optional

import httpx
from exchange.content import Text, ToolResult, ToolUse
from exchange.message import Message
from exchange.tool import Tool
from tenacity import retry_if_exception


def retry_if_status(codes: Optional[list[int]] = None, above: Optional[int] = None) -> callable:
    codes = codes or []

    def predicate(exc: BaseException) -> bool:
        if isinstance(exc, httpx.HTTPStatusError):
            if exc.response.status_code in codes:
                return True
            if above and exc.response.status_code >= above:
                return True
        return False

    return retry_if_exception(predicate)


def raise_for_status(response: httpx.Response) -> httpx.Response:
    """Raise with reason text."""
    try:
        response.raise_for_status()
        return response
    except httpx.HTTPStatusError as e:
        response.read()
        if response.text:
            raise httpx.HTTPStatusError(f"{e}\n{response.text}", request=e.request, response=e.response)
        else:
            raise e


def encode_image(image_path: str) -> str:
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")


def messages_to_openai_spec(messages: list[Message]) -> list[dict[str, any]]:
    messages_spec = []
    for message in messages:
        converted = {"role": message.role}
        output = []
        for content in message.content:
            if isinstance(content, Text):
                converted["content"] = content.text
            elif isinstance(content, ToolUse):
                sanitized_name = re.sub(r"[^a-zA-Z0-9_-]", "_", content.name)
                converted.setdefault("tool_calls", []).append(
                    {
                        "id": content.id,
                        "type": "function",
                        "function": {
                            "name": sanitized_name,
                            "arguments": json.dumps(content.parameters),
                        },
                    }
                )
            elif isinstance(content, ToolResult):
                if content.output.startswith('"image:'):
                    image_path = content.output.replace('"image:', "").replace('"', "")
                    output.append(
                        {
                            "role": "tool",
                            "content": [
                                {
                                    "type": "text",
                                    "text": "This tool result included an image that is uploaded in the next message.",
                                },
                            ],
                            "tool_call_id": content.tool_use_id,
                        }
                    )
                    # Note: it is possible to only do this when message == messages[-1]
                    # but it doesn't seem to hurt too much with tokens to keep this.
                    output.append(
                        {
                            "role": "user",
                            "content": [
                                {
                                    "type": "image_url",
                                    "image_url": {"url": f"data:image/jpeg;base64,{encode_image(image_path)}"},
                                }
                            ],
                        }
                    )

                else:
                    output.append(
                        {
                            "role": "tool",
                            "content": content.output,
                            "tool_call_id": content.tool_use_id,
                        }
                    )

        if "content" in converted or "tool_calls" in converted:
            output = [converted] + output
        messages_spec.extend(output)
    return messages_spec


def tools_to_openai_spec(tools: tuple[Tool, ...]) -> dict[str, any]:
    tools_names = set()
    result = []
    for tool in tools:
        if tool.name in tools_names:
            # we should never allow duplicate tools
            raise ValueError(f"Duplicate tool name: {tool.name}")
        result.append(
            {
                "type": "function",
                "function": {
                    "name": tool.name,
                    "description": tool.description,
                    "parameters": tool.parameters,
                },
            }
        )
        tools_names.add(tool.name)
    return result


def openai_response_to_message(response: dict) -> Message:
    original = response["choices"][0]["message"]
    content = []
    text = original.get("content")
    if text:
        content.append(Text(text=text))

    tool_calls = original.get("tool_calls")
    if tool_calls:
        for tool_call in tool_calls:
            try:
                function_name = tool_call["function"]["name"]
                # We occasionally see the model generate an invalid function name
                # sending this back to openai raises a validation error
                if not re.match(r"^[a-zA-Z0-9_-]+$", function_name):
                    content.append(
                        ToolUse(
                            id=tool_call["id"],
                            name=function_name,
                            parameters=tool_call["function"]["arguments"],
                            is_error=True,
                            error_message=f"The provided function name '{function_name}' had invalid characters, it must match this regex [a-zA-Z0-9_-]+",  # noqa: E501
                        )
                    )
                else:
                    content.append(
                        ToolUse(
                            id=tool_call["id"],
                            name=function_name,
                            parameters=json.loads(tool_call["function"]["arguments"]),
                        )
                    )
            except json.JSONDecodeError:
                content.append(
                    ToolUse(
                        id=tool_call["id"],
                        name=tool_call["function"]["name"],
                        parameters=tool_call["function"]["arguments"],
                        is_error=True,
                        error_message=f"Could not interpret tool use parameters for id {tool_call['id']}: {tool_call['function']['arguments']}",  # noqa: E501
                    )
                )

    return Message(role="assistant", content=content)


def openai_single_message_context_length_exceeded(error_dict: dict) -> None:
    code = error_dict.get("code")
    if code == "context_length_exceeded" or code == "string_above_max_length":
        raise InitialMessageTooLargeError(f"Input message too long. Message: {error_dict.get('message')}")


class InitialMessageTooLargeError(Exception):
    """Custom error raised when the first input message in an exchange is too large."""

    pass



class InvalidChoiceError(Exception):
    def __init__(self, attribute_name: str, attribute_value: str, available_values: list[str]) -> None:
        self.attribute_name = attribute_name
        self.attribute_value = attribute_value
        self.available_values = available_values
        self.message = (
            f"Unknown {attribute_name}: {attribute_value}."
            + f" Available {attribute_name}s: {', '.join(available_values)}"
        )
        super().__init__(self.message)



from typing import Optional

import json
from attrs import define, asdict


CONTENT_TYPES = {}


class Content:
    def __init_subclass__(cls, **kwargs: dict[str, any]) -> None:
        super().__init_subclass__(**kwargs)
        CONTENT_TYPES[cls.__name__] = cls

    def to_dict(self) -> dict[str, any]:
        data = asdict(self, recurse=True)
        data["type"] = self.__class__.__name__
        return data


@define
class Text(Content):
    text: str

    @property
    def summary(self) -> str:
        return "content:text\n" + self.text


@define
class ToolUse(Content):
    id: str
    name: str
    parameters: any
    is_error: bool = False
    error_message: Optional[str] = None

    @property
    def summary(self) -> str:
        return f"content:tool_use:{self.name}\nparameters:{json.dumps(self.parameters)}"


@define
class ToolResult(Content):
    tool_use_id: str
    output: str
    is_error: bool = False

    @property
    def summary(self) -> str:
        return f"content:tool_result:error={self.is_error}\noutput:{self.output}"



import pytest
from unittest.mock import patch, MagicMock
from exchange.langfuse_wrapper import observe_wrapper


@pytest.fixture
def mock_langfuse_context():
    with patch("exchange.langfuse_wrapper.langfuse_context") as mock:
        yield mock


@patch("exchange.langfuse_wrapper.HAS_LANGFUSE_CREDENTIALS", True)
def test_function_is_wrapped(mock_langfuse_context):
    mock_observe = MagicMock(side_effect=lambda *args, **kwargs: lambda fn: fn)
    mock_langfuse_context.observe = mock_observe

    def original_function(x: int, y: int) -> int:
        return x + y

    # test function before we decorate it with
    # @observe_wrapper("arg1", kwarg1="kwarg1")
    assert not hasattr(original_function, "__wrapped__")

    # ensure we args get passed along (e.g. @observe(capture_input=False, capture_output=False))
    decorated_function = observe_wrapper("arg1", kwarg1="kwarg1")(original_function)
    assert hasattr(decorated_function, "__wrapped__")
    assert decorated_function.__wrapped__ is original_function, "Function is not properly wrapped"

    assert decorated_function(2, 3) == 5
    mock_observe.assert_called_once()
    mock_observe.assert_called_with("arg1", kwarg1="kwarg1")


@patch("exchange.langfuse_wrapper.HAS_LANGFUSE_CREDENTIALS", False)
def test_function_is_not_wrapped(mock_langfuse_context):
    mock_observe = MagicMock(return_value=lambda f: f)
    mock_langfuse_context.observe = mock_observe

    @observe_wrapper("arg1", kwarg1="kwarg1")
    def hello() -> str:
        return "Hello"

    assert not hasattr(hello, "__wrapped__")
    assert hello() == "Hello"

    mock_observe.assert_not_called()



import json
import os
import re

import pytest
import yaml

from exchange import Message, ToolUse, ToolResult, Tool
from exchange.providers import Usage, Provider

from tests.conftest import read_file

OPENAI_API_KEY = "test_openai_api_key"
OPENAI_ORG_ID = "test_openai_org_key"
OPENAI_PROJECT_ID = "test_openai_project_id"


@pytest.fixture
def default_openai_env(monkeypatch):
    """
    This fixture prevents OpenAIProvider.from_env() from erring on missing
    environment variables.

    When running VCR tests for the first time or after deleting a cassette
    recording, set required environment variables, so that real requests don't
    fail. Subsequent runs use the recorded data, so don't need them.
    """
    if "OPENAI_API_KEY" not in os.environ:
        monkeypatch.setenv("OPENAI_API_KEY", OPENAI_API_KEY)


AZURE_ENDPOINT = "https://test.openai.azure.com"
AZURE_DEPLOYMENT_NAME = "test-azure-deployment"
AZURE_API_VERSION = "2024-05-01-preview"
AZURE_API_KEY = "test_azure_api_key"


@pytest.fixture
def default_azure_env(monkeypatch):
    """
    This fixture prevents AzureProvider.from_env() from erring on missing
    environment variables.

    When running VCR tests for the first time or after deleting a cassette
    recording, set required environment variables, so that real requests don't
    fail. Subsequent runs use the recorded data, so don't need them.
    """
    if "AZURE_CHAT_COMPLETIONS_HOST_NAME" not in os.environ:
        monkeypatch.setenv("AZURE_CHAT_COMPLETIONS_HOST_NAME", AZURE_ENDPOINT)
    if "AZURE_CHAT_COMPLETIONS_DEPLOYMENT_NAME" not in os.environ:
        monkeypatch.setenv("AZURE_CHAT_COMPLETIONS_DEPLOYMENT_NAME", AZURE_DEPLOYMENT_NAME)
    if "AZURE_CHAT_COMPLETIONS_DEPLOYMENT_API_VERSION" not in os.environ:
        monkeypatch.setenv("AZURE_CHAT_COMPLETIONS_DEPLOYMENT_API_VERSION", AZURE_API_VERSION)
    if "AZURE_CHAT_COMPLETIONS_KEY" not in os.environ:
        monkeypatch.setenv("AZURE_CHAT_COMPLETIONS_KEY", AZURE_API_KEY)


GOOGLE_API_KEY = "test_google_api_key"


@pytest.fixture
def default_google_env(monkeypatch):
    """
    This fixture prevents GoogleProvider.from_env() from erring on missing
    environment variables.

    When running VCR tests for the first time or after deleting a cassette
    recording, set required environment variables, so that real requests don't
    fail. Subsequent runs use the recorded data, so don't need them.
    """
    if "GOOGLE_API_KEY" not in os.environ:
        monkeypatch.setenv("GOOGLE_API_KEY", GOOGLE_API_KEY)


class LiteralBlockScalar(str):
    """Formats the string as a literal block scalar, preserving whitespace and
    without interpreting escape characters"""

    pass


def literal_block_scalar_presenter(dumper, data):
    """Represents a scalar string as a literal block, via '|' syntax"""
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")


yaml.add_representer(LiteralBlockScalar, literal_block_scalar_presenter)


def process_string_value(string_value):
    """Pretty-prints JSON or returns long strings as a LiteralString"""
    try:
        json_data = json.loads(string_value)
        return LiteralBlockScalar(json.dumps(json_data, indent=2))
    except (ValueError, TypeError):
        if len(string_value) > 80:
            return LiteralBlockScalar(string_value)
    return string_value


def convert_body_to_literal(data):
    """Searches the data for body strings, attempting to pretty-print JSON"""
    if isinstance(data, dict):
        for key, value in data.items():
            # Handle response body case (e.g., response.body.string)
            if key == "body" and isinstance(value, dict) and "string" in value:
                value["string"] = process_string_value(value["string"])

            # Handle request body case (e.g., request.body)
            elif key == "body" and isinstance(value, str):
                data[key] = process_string_value(value)

            else:
                convert_body_to_literal(value)

    elif isinstance(data, list):
        for i, item in enumerate(data):
            data[i] = convert_body_to_literal(item)

    return data


class PrettyPrintJSONBody:
    """This makes request and response body recordings more readable."""

    @staticmethod
    def serialize(cassette_dict):
        cassette_dict = convert_body_to_literal(cassette_dict)
        return yaml.dump(cassette_dict, default_flow_style=False, allow_unicode=True)

    @staticmethod
    def deserialize(cassette_string):
        return yaml.load(cassette_string, Loader=yaml.Loader)


@pytest.fixture(scope="module")
def vcr(vcr):
    vcr.register_serializer("yaml", PrettyPrintJSONBody)
    return vcr


@pytest.fixture(scope="module")
def vcr_config():
    """
    This scrubs sensitive data and gunzips bodies when in recording mode.

    Without this, you would leak cookies and auth tokens in the cassettes.
    Also, depending on the request, some responses would be binary encoded
    while others plain json. This ensures all bodies are human-readable.
    """
    return {
        "decode_compressed_response": True,
        "filter_headers": [
            ("authorization", "Bearer " + OPENAI_API_KEY),
            ("openai-organization", OPENAI_ORG_ID),
            ("openai-project", OPENAI_PROJECT_ID),
            ("cookie", None),
        ],
        "before_record_request": scrub_request_url,
        "before_record_response": scrub_response_headers,
    }


def scrub_request_url(request):
    """
    This scrubs sensitive request data in provider-specific way. Note that headers
    are case-sensitive!
    """
    if "openai.azure.com" in request.uri:
        request.uri = re.sub(r"https://[^/]+", AZURE_ENDPOINT, request.uri)
        request.uri = re.sub(r"/deployments/[^/]+", f"/deployments/{AZURE_DEPLOYMENT_NAME}", request.uri)
        request.headers["host"] = AZURE_ENDPOINT.replace("https://", "")
        request.headers["api-key"] = AZURE_API_KEY
    elif "generativelanguage.googleapis.com" in request.uri:
        request.uri = re.sub(r"([?&])key=[^&]+", r"\1key=" + GOOGLE_API_KEY, request.uri)

    return request


def scrub_response_headers(response):
    """
    This scrubs sensitive response headers. Note they are case-sensitive!
    """
    if "openai-organization" in response["headers"]:
        response["headers"]["openai-organization"] = OPENAI_ORG_ID
    if "Set-Cookie" in response["headers"]:
        response["headers"]["Set-Cookie"] = "test_set_cookie"
    return response


def complete(provider_cls: type[Provider], model: str, **kwargs) -> tuple[Message, Usage]:
    provider = provider_cls.from_env()
    system = "You are a helpful assistant."
    messages = [Message.user("Hello")]
    return provider.complete(model=model, system=system, messages=messages, tools=(), **kwargs)


def tools(provider_cls: type[Provider], model: str, **kwargs) -> tuple[Message, Usage]:
    provider = provider_cls.from_env()
    system = "You are a helpful assistant. Expect to need to read a file using read_file."
    messages = [Message.user("What are the contents of this file? test.txt")]
    return provider.complete(
        model=model, system=system, messages=messages, tools=(Tool.from_function(read_file),), **kwargs
    )


def vision(provider_cls: type[Provider], model: str, **kwargs) -> tuple[Message, Usage]:
    provider = provider_cls.from_env()
    system = "You are a helpful assistant."
    messages = [
        Message.user("What does the first entry in the menu say?"),
        Message(
            role="assistant",
            content=[ToolUse(id="xyz", name="screenshot", parameters={})],
        ),
        Message(
            role="user",
            content=[ToolResult(tool_use_id="xyz", output='"image:tests/test_image.png"')],
        ),
    ]
    return provider.complete(model=model, system=system, messages=messages, tools=(), **kwargs)



import httpx
import os

from exchange.providers import OpenAiProvider


class AzureProvider(OpenAiProvider):
    """Provides chat completions for models hosted by the Azure OpenAI Service."""

    PROVIDER_NAME = "azure"
    REQUIRED_ENV_VARS = [
        "AZURE_CHAT_COMPLETIONS_HOST_NAME",
        "AZURE_CHAT_COMPLETIONS_DEPLOYMENT_NAME",
        "AZURE_CHAT_COMPLETIONS_DEPLOYMENT_API_VERSION",
        "AZURE_CHAT_COMPLETIONS_KEY",
    ]

    def __init__(self, client: httpx.Client) -> None:
        super().__init__(client)

    @classmethod
    def from_env(cls: type["AzureProvider"]) -> "AzureProvider":
        cls.check_env_vars()
        url = os.environ.get("AZURE_CHAT_COMPLETIONS_HOST_NAME")
        deployment_name = os.environ.get("AZURE_CHAT_COMPLETIONS_DEPLOYMENT_NAME")
        api_version = os.environ.get("AZURE_CHAT_COMPLETIONS_DEPLOYMENT_API_VERSION")
        key = os.environ.get("AZURE_CHAT_COMPLETIONS_KEY")

        # format the url host/"openai/deployments/" + deployment_name + "/?api-version=" + api_version
        url = f"{url}/openai/deployments/{deployment_name}/"
        client = httpx.Client(
            base_url=url,
            headers={"api-key": key, "Content-Type": "application/json"},
            params={"api-version": api_version},
            timeout=httpx.Timeout(60 * 10),
        )
        return cls(client)



import os
from unittest.mock import patch

import httpx
import pytest
from exchange import Message, Text
from exchange.content import ToolResult, ToolUse
from exchange.providers.anthropic import AnthropicProvider
from exchange.providers.base import MissingProviderEnvVariableError
from exchange.tool import Tool


def example_fn(param: str) -> None:
    """
    Testing function.

    Args:
        param (str): Description of param1
    """
    pass


@pytest.fixture
@patch.dict(os.environ, {"ANTHROPIC_API_KEY": "test_api_key"})
def anthropic_provider():
    return AnthropicProvider.from_env()


def test_from_env_throw_error_when_missing_api_key():
    with patch.dict(os.environ, {}, clear=True):
        with pytest.raises(MissingProviderEnvVariableError) as context:
            AnthropicProvider.from_env()
        assert context.value.provider == "anthropic"
        assert context.value.env_variable == "ANTHROPIC_API_KEY"
        assert context.value.message == "Missing environment variable: ANTHROPIC_API_KEY for provider anthropic."


def test_anthropic_response_to_text_message() -> None:
    response = {
        "content": [{"type": "text", "text": "Hello from Claude!"}],
    }
    message = AnthropicProvider.anthropic_response_to_message(response)
    assert message.content[0].text == "Hello from Claude!"


def test_anthropic_response_to_tool_use_message() -> None:
    response = {
        "content": [
            {
                "type": "tool_use",
                "id": "1",
                "name": "example_fn",
                "input": {"param": "value"},
            }
        ],
    }
    message = AnthropicProvider.anthropic_response_to_message(response)
    assert message.content[0].id == "1"
    assert message.content[0].name == "example_fn"
    assert message.content[0].parameters == {"param": "value"}


def test_tools_to_anthropic_spec() -> None:
    tools = (Tool.from_function(example_fn),)
    expected_spec = [
        {
            "name": "example_fn",
            "description": "Testing function.",
            "input_schema": {
                "type": "object",
                "properties": {"param": {"type": "string", "description": "Description of param1"}},
                "required": ["param"],
            },
        }
    ]
    result = AnthropicProvider.tools_to_anthropic_spec(tools)
    assert result == expected_spec


def test_message_text_to_anthropic_spec() -> None:
    messages = [Message.user("Hello, Claude")]
    expected_spec = [
        {
            "role": "user",
            "content": [{"type": "text", "text": "Hello, Claude"}],
        }
    ]
    result = AnthropicProvider.messages_to_anthropic_spec(messages)
    assert result == expected_spec


def test_messages_to_anthropic_spec() -> None:
    messages = [
        Message(role="user", content=[Text("Hello, Claude")]),
        Message(
            role="assistant",
            content=[ToolUse(id="1", name="example_fn", parameters={"param": "value"})],
        ),
        Message(role="user", content=[ToolResult(tool_use_id="1", output="Result")]),
    ]
    actual_spec = AnthropicProvider.messages_to_anthropic_spec(messages)
    #  !=
    expected_spec = [
        {"role": "user", "content": [{"type": "text", "text": "Hello, Claude"}]},
        {
            "role": "assistant",
            "content": [
                {
                    "type": "tool_use",
                    "id": "1",
                    "name": "example_fn",
                    "input": {"param": "value"},
                }
            ],
        },
        {
            "role": "user",
            "content": [{"type": "tool_result", "tool_use_id": "1", "content": "Result"}],
        },
    ]
    assert actual_spec == expected_spec


@patch("httpx.Client.post")
@patch("logging.warning")
@patch("logging.error")
def test_anthropic_completion(mock_error, mock_warning, mock_post, anthropic_provider):
    mock_response = {
        "content": [{"type": "text", "text": "Hello from Claude!"}],
        "usage": {"input_tokens": 10, "output_tokens": 25},
    }

    # First attempts fail with status code 429, 2nd succeeds
    def create_response(status_code, json_data=None):
        response = httpx.Response(status_code)
        response._content = httpx._content.json_dumps(json_data or {}).encode()
        response._request = httpx.Request("POST", "https://api.anthropic.com/v1/messages")
        return response

    mock_post.side_effect = [
        create_response(429),  # 1st attempt
        create_response(200, mock_response),  # Final success
    ]

    model = "claude-3-5-sonnet-20240620"
    system = "You are a helpful assistant."
    messages = [Message.user("Hello, Claude")]

    reply_message, reply_usage = anthropic_provider.complete(model=model, system=system, messages=messages)

    assert reply_message.content == [Text("Hello from Claude!")]
    assert reply_usage.total_tokens == 35
    assert mock_post.call_count == 2
    mock_post.assert_any_call(
        "https://api.anthropic.com/v1/messages",
        json={
            "system": system,
            "model": model,
            "max_tokens": 4096,
            "messages": [
                *[
                    {
                        "role": msg.role,
                        "content": [{"type": "text", "text": msg.content[0].text}],
                    }
                    for msg in messages
                ],
            ],
        },
    )


@pytest.mark.integration
def test_anthropic_integration():
    provider = AnthropicProvider.from_env()
    model = "claude-3-5-sonnet-20240620"  # updated model to a known valid model
    system = "You are a helpful assistant."
    messages = [Message.user("Hello, Claude")]

    # Run the completion
    reply = provider.complete(model=model, system=system, messages=messages)

    assert reply[0].content is not None
    print("Completion content from Anthropic:", reply[0].content)



from exchange.providers.base import MissingProviderEnvVariableError


def test_missing_provider_env_variable_error_without_instructions_url():
    env_variable = "API_KEY"
    provider = "TestProvider"
    error = MissingProviderEnvVariableError(env_variable, provider)

    assert error.env_variable == env_variable
    assert error.provider == provider
    assert error.instructions_url is None
    assert error.message == "Missing environment variable: API_KEY for provider TestProvider."


def test_missing_provider_env_variable_error_with_instructions_url():
    env_variable = "API_KEY"
    provider = "TestProvider"
    instructions_url = "http://example.com/instructions"
    error = MissingProviderEnvVariableError(env_variable, provider, instructions_url)

    assert error.env_variable == env_variable
    assert error.provider == provider
    assert error.instructions_url == instructions_url
    assert error.message == (
        "Missing environment variable: API_KEY for provider TestProvider.\n"
        "Please see http://example.com/instructions for instructions"
    )



import os
from unittest.mock import patch

import pytest

from exchange import Text, ToolUse
from exchange.providers.azure import AzureProvider
from exchange.providers.base import MissingProviderEnvVariableError
from .conftest import complete, tools

AZURE_MODEL = os.getenv("AZURE_MODEL", "gpt-4o-mini")


@pytest.mark.parametrize(
    "env_var_name",
    [
        "AZURE_CHAT_COMPLETIONS_HOST_NAME",
        "AZURE_CHAT_COMPLETIONS_DEPLOYMENT_NAME",
        "AZURE_CHAT_COMPLETIONS_DEPLOYMENT_API_VERSION",
        "AZURE_CHAT_COMPLETIONS_KEY",
    ],
)
def test_from_env_throw_error_when_missing_env_var(env_var_name):
    with patch.dict(
        os.environ,
        {
            "AZURE_CHAT_COMPLETIONS_HOST_NAME": "test_host_name",
            "AZURE_CHAT_COMPLETIONS_DEPLOYMENT_NAME": "test_deployment_name",
            "AZURE_CHAT_COMPLETIONS_DEPLOYMENT_API_VERSION": "test_api_version",
            "AZURE_CHAT_COMPLETIONS_KEY": "test_api_key",
        },
        clear=True,
    ):
        os.environ.pop(env_var_name)
        with pytest.raises(MissingProviderEnvVariableError) as context:
            AzureProvider.from_env()
        assert context.value.provider == "azure"
        assert context.value.env_variable == env_var_name
        assert context.value.message == f"Missing environment variable: {env_var_name} for provider azure."


@pytest.mark.vcr()
def test_azure_complete(default_azure_env):
    reply_message, reply_usage = complete(AzureProvider, AZURE_MODEL)

    assert reply_message.content == [Text("Hello! How can I assist you today?\n")]
    assert reply_usage.total_tokens == 27


@pytest.mark.integration
def test_azure_complete_integration():
    reply = complete(AzureProvider, AZURE_MODEL)

    assert reply[0].content is not None
    print("Completion content from Azure:", reply[0].content)


@pytest.mark.vcr()
def test_azure_tools(default_azure_env):
    reply_message, reply_usage = tools(AzureProvider, AZURE_MODEL)

    tool_use = reply_message.content[0]
    assert isinstance(tool_use, ToolUse), f"Expected ToolUse, but was {type(tool_use).__name__}"
    assert tool_use.id == "call_Dn0idyNSdmHSYpsql3EbFH9L"
    assert tool_use.name == "read_file"
    assert tool_use.parameters == {"filename": "test.txt"}
    assert reply_usage.total_tokens == 125


@pytest.mark.integration
def test_azure_tools_integration():
    reply = tools(AzureProvider, AZURE_MODEL)

    tool_use = reply[0].content[0]
    assert isinstance(tool_use, ToolUse), f"Expected ToolUse, but was {type(tool_use).__name__}"
    assert tool_use.id is not None
    assert tool_use.name == "read_file"
    assert tool_use.parameters == {"filename": "test.txt"}



import os
from unittest.mock import patch

import pytest
from exchange import Message, Text
from exchange.content import ToolResult, ToolUse
from exchange.providers.base import MissingProviderEnvVariableError
from exchange.providers.google import GoogleProvider
from exchange.tool import Tool
from .conftest import complete, tools

GOOGLE_MODEL = os.getenv("GOOGLE_MODEL", "gemini-1.5-flash")


def example_fn(param: str) -> None:
    """
    Testing function.

    Args:
        param (str): Description of param1
    """
    pass


def test_from_env_throw_error_when_missing_api_key():
    with patch.dict(os.environ, {}, clear=True):
        with pytest.raises(MissingProviderEnvVariableError) as context:
            GoogleProvider.from_env()
        assert context.value.provider == "google"
        assert context.value.env_variable == "GOOGLE_API_KEY"
        assert "Missing environment variable: GOOGLE_API_KEY for provider google" in context.value.message
        assert "https://ai.google.dev/gemini-api/docs/api-key" in context.value.message


def test_google_response_to_text_message() -> None:
    response = {"candidates": [{"content": {"parts": [{"text": "Hello from Gemini!"}], "role": "model"}}]}
    message = GoogleProvider.google_response_to_message(response)
    assert message.content[0].text == "Hello from Gemini!"


def test_google_response_to_tool_use_message() -> None:
    response = {
        "candidates": [
            {
                "content": {
                    "parts": [{"functionCall": {"name": "example_fn", "args": {"param": "value"}}}],
                    "role": "model",
                }
            }
        ]
    }

    message = GoogleProvider.google_response_to_message(response)
    assert message.content[0].name == "example_fn"
    assert message.content[0].parameters == {"param": "value"}


def test_tools_to_google_spec() -> None:
    tools = (Tool.from_function(example_fn),)
    expected_spec = {
        "functionDeclarations": [
            {
                "name": "example_fn",
                "description": "Testing function.",
                "parameters": {
                    "type": "object",
                    "properties": {"param": {"type": "string", "description": "Description of param1"}},
                    "required": ["param"],
                },
            }
        ]
    }
    result = GoogleProvider.tools_to_google_spec(tools)
    assert result == expected_spec


def test_message_text_to_google_spec() -> None:
    messages = [Message.user("Hello, Gemini")]
    expected_spec = [{"role": "user", "parts": [{"text": "Hello, Gemini"}]}]
    result = GoogleProvider.messages_to_google_spec(messages)
    assert result == expected_spec


def test_messages_to_google_spec() -> None:
    messages = [
        Message(role="user", content=[Text("Hello, Gemini")]),
        Message(
            role="assistant",
            content=[ToolUse(id="1", name="example_fn", parameters={"param": "value"})],
        ),
        Message(role="user", content=[ToolResult(tool_use_id="1", output="Result")]),
    ]
    actual_spec = GoogleProvider.messages_to_google_spec(messages)
    #  !=
    expected_spec = [
        {"role": "user", "parts": [{"text": "Hello, Gemini"}]},
        {"role": "model", "parts": [{"functionCall": {"name": "example_fn", "args": {"param": "value"}}}]},
        {"role": "user", "parts": [{"functionResponse": {"name": "1", "response": {"content": "Result"}}}]},
    ]

    assert actual_spec == expected_spec


@pytest.mark.vcr()
def test_google_complete(default_google_env):
    reply_message, reply_usage = complete(GoogleProvider, GOOGLE_MODEL)

    assert reply_message.content == [Text("Hello! 👋  What can I do for you today? 😊 \n")]
    assert reply_usage.total_tokens == 21


@pytest.mark.integration
def test_google_complete_integration():
    reply = complete(GoogleProvider, GOOGLE_MODEL)

    assert reply[0].content is not None
    print("Completion content from Google:", reply[0].content)


@pytest.mark.vcr()
def test_google_tools(default_google_env):
    reply_message, reply_usage = tools(GoogleProvider, GOOGLE_MODEL)

    tool_use = reply_message.content[0]
    assert isinstance(tool_use, ToolUse), f"Expected ToolUse, but was {type(tool_use).__name__}"
    assert tool_use.id == "read_file"
    assert tool_use.name == "read_file"
    assert tool_use.parameters == {"filename": "test.txt"}
    assert reply_usage.total_tokens == 118


@pytest.mark.integration
def test_google_tools_integration():
    reply = tools(GoogleProvider, GOOGLE_MODEL)

    tool_use = reply[0].content[0]
    assert isinstance(tool_use, ToolUse), f"Expected ToolUse, but was {type(tool_use).__name__}"
    assert tool_use.id is not None
    assert tool_use.name == "read_file"
    assert tool_use.parameters == {"filename": "test.txt"}



from functools import cache

from exchange.invalid_choice_error import InvalidChoiceError
from exchange.moderators.base import Moderator
from exchange.utils import load_plugins
from exchange.moderators.passive import PassiveModerator  # noqa
from exchange.moderators.truncate import ContextTruncate  # noqa
from exchange.moderators.summarizer import ContextSummarizer  # noqa


@cache
def get_moderator(name: str) -> type[Moderator]:
    moderators = load_plugins(group="exchange.moderator")
    if name not in moderators:
        raise InvalidChoiceError("moderator", name, moderators.keys())
    return moderators[name]



"""
Langfuse Integration Module

This module provides integration with Langfuse, a tool for monitoring and tracing LLM applications.

Usage:
    Import this module to enable Langfuse integration.
    It automatically checks for Langfuse credentials in the .env.langfuse file and for a running Langfuse server.
    If these are found, it will set up the necessary client and context for tracing.

Note:
    Run setup_langfuse.sh which automates the steps for running local Langfuse.
"""

import os
from typing import Callable
from dotenv import load_dotenv
from langfuse.decorators import langfuse_context
import sys
from io import StringIO
from pathlib import Path
from functools import wraps  # Add this import


def find_package_root(start_path: Path, marker_file: str = "pyproject.toml") -> Path:
    while start_path != start_path.parent:
        if (start_path / marker_file).exists():
            return start_path
        start_path = start_path.parent
    return None


def auth_check() -> bool:
    # Temporarily redirect stdout and stderr to suppress print statements from Langfuse
    temp_stderr = StringIO()
    sys.stderr = temp_stderr

    # Load environment variables
    load_dotenv(LANGFUSE_ENV_FILE, override=True)

    auth_val = langfuse_context.auth_check()

    # Restore stderr
    sys.stderr = sys.__stderr__
    return auth_val


CURRENT_DIR = Path(__file__).parent
PACKAGE_ROOT = find_package_root(CURRENT_DIR)

LANGFUSE_ENV_FILE = os.path.join(PACKAGE_ROOT, ".env.langfuse.local")
HAS_LANGFUSE_CREDENTIALS = False
load_dotenv(LANGFUSE_ENV_FILE, override=True)

HAS_LANGFUSE_CREDENTIALS = auth_check()


def observe_wrapper(*args, **kwargs) -> Callable:  # noqa
    """
    A decorator that wraps a function with Langfuse context observation if credentials are available.

    If Langfuse credentials were found, the function will be wrapped with Langfuse's observe method.
    Otherwise, the function will be returned as-is.

    Args:
        *args: Positional arguments to pass to langfuse_context.observe.
        **kwargs: Keyword arguments to pass to langfuse_context.observe.

    Returns:
        Callable: The wrapped function if credentials are available, otherwise the original function.
    """

    def _wrapper(fn: Callable) -> Callable:
        if HAS_LANGFUSE_CREDENTIALS:

            @wraps(fn)
            def wrapped_fn(*fargs, **fkwargs):  # noqa
                return langfuse_context.observe(*args, **kwargs)(fn)(*fargs, **fkwargs)

            return wrapped_fn
        else:
            return fn

    return _wrapper



from exchange.invalid_choice_error import InvalidChoiceError
from exchange.moderators import get_moderator
import pytest


def test_get_moderator():
    moderator = get_moderator("truncate")
    assert moderator.__name__ == "ContextTruncate"


def test_get_moderator_raise_error_for_unknown_moderator():
    with pytest.raises(InvalidChoiceError) as error:
        get_moderator("nonexistent")
    assert error.value.attribute_name == "moderator"
    assert error.value.attribute_value == "nonexistent"
    assert "truncate" in error.value.available_values
    assert "truncate" in error.value.message



import os
from abc import ABC, abstractmethod
from attrs import define, field
from typing import Optional

from exchange.message import Message
from exchange.tool import Tool


@define(hash=True)
class Usage:
    input_tokens: int = field(factory=None)
    output_tokens: int = field(default=None)
    total_tokens: int = field(default=None)


class Provider(ABC):
    PROVIDER_NAME: str
    REQUIRED_ENV_VARS: list[str] = []

    @classmethod
    def from_env(cls: type["Provider"]) -> "Provider":
        return cls()

    @classmethod
    def check_env_vars(cls: type["Provider"], instructions_url: Optional[str] = None) -> None:
        for env_var in cls.REQUIRED_ENV_VARS:
            if env_var not in os.environ:
                raise MissingProviderEnvVariableError(env_var, cls.PROVIDER_NAME, instructions_url)

    @abstractmethod
    def complete(
        self,
        model: str,
        system: str,
        messages: list[Message],
        tools: tuple[Tool, ...],
        **kwargs: dict[str, any],
    ) -> tuple[Message, Usage]:
        """Generate the next message using the specified model"""
        pass


class MissingProviderEnvVariableError(Exception):
    def __init__(self, env_variable: str, provider: str, instructions_url: Optional[str] = None) -> None:
        self.env_variable = env_variable
        self.provider = provider
        self.instructions_url = instructions_url
        self.message = f"Missing environment variable: {env_variable} for provider {provider}."
        if instructions_url:
            self.message += f"\nPlease see {instructions_url} for instructions"
        super().__init__(self.message)
